"""
SwFTP is an FTP and SFTP interface for Openstack Swift

See COPYING for license information.
"""
from swftp.utils import VERSION
__title__ = 'swftp'
__version__ = VERSION
__author__ = 'SoftLayer Technologies, Inc.'
__license__ = 'MIT'
__copyright__ = 'Copyright 2013 SoftLayer Technologies, Inc.'
